from typing import List, Optional, Union, cast

import requests
from eip712_structs import make_domain
from eth_account import Account
from eth_account.messages import encode_defunct
from eth_typing import AnyAddress, ChecksumAddress, HexStr
from hexbytes import HexBytes
from web3 import Web3

from gnosis.eth import EthereumNetwork, EthereumNetworkNotSupported

from .order import Order, OrderKind

try:
    from typing import TypedDict  # pylint: disable=no-name-in-module
except ImportError:
    from typing_extensions import TypedDict

try:
    from functools import cache
except ImportError:
    from functools import lru_cache

    cache = lru_cache(maxsize=None)


class TradeResponse(TypedDict):
    blockNumber: int
    logIndex: int
    orderUid: HexStr
    buyAmount: str  # Stringified int
    sellAmount: str  # Stringified int
    sellAmountBeforeFees: str  # Stringified int
    owner: AnyAddress  # Not checksummed
    buyToken: AnyAddress
    sellToken: AnyAddress
    txHash: HexStr


class AmountResponse(TypedDict):
    amount: str
    token: AnyAddress


class ErrorResponse(TypedDict):
    error_type: str
    description: str


class GnosisProtocolAPI:
    """
    Client for GnosisProtocol API. More info: https://docs.cowswap.exchange/
    """

    settlement_contract_addresses = {
        EthereumNetwork.MAINNET: "0x9008D19f58AAbD9eD0D60971565AA8510560ab41",
        EthereumNetwork.RINKEBY: "0x9008D19f58AAbD9eD0D60971565AA8510560ab41",
        EthereumNetwork.XDAI: "0x9008D19f58AAbD9eD0D60971565AA8510560ab41",
    }

    api_base_urls = {
        EthereumNetwork.MAINNET: "https://protocol-mainnet.gnosis.io/api/v1/",
        EthereumNetwork.RINKEBY: "https://protocol-rinkeby.gnosis.io/api/v1/",
        EthereumNetwork.XDAI: "https://protocol-xdai.gnosis.io/api/v1/",
    }

    def __init__(self, ethereum_network: EthereumNetwork):
        self.network = ethereum_network
        if self.network not in self.api_base_urls:
            raise EthereumNetworkNotSupported(
                f"{self.network.name} network not supported by Gnosis Protocol"
            )
        self.domain_separator = self.build_domain_separator(self.network)
        self.base_url = self.api_base_urls[self.network]

    @classmethod
    def build_domain_separator(cls, ethereum_network: EthereumNetwork):
        return make_domain(
            name="Gnosis Protocol",
            version="v2",
            chainId=str(ethereum_network.value),
            verifyingContract=cls.settlement_contract_addresses[ethereum_network],
        )

    def get_fee(self, order: Order) -> int:
        if order["kind"] == "sell":
            amount = order["sellAmount"]
        else:
            amount = order["buyAmount"]
        url = (
            self.base_url
            + f'fee/?sellToken={order["sellToken"]}&buyToken={order["buyToken"]}'
            f'&amount={amount}&kind={order["kind"]}'
        )
        result = requests.get(url).json()
        if "amount" in result:
            return int(result["amount"])
        else:
            return 0

    def place_order(
        self, order: Order, private_key: HexStr
    ) -> Union[HexStr, ErrorResponse]:
        """
        Place order. If `feeAmount=0` in Order it will be calculated calling `get_fee(order)`

        :return: UUID for the order as an hex hash
        """
        assert (
            order["buyAmount"] and order["sellAmount"]
        ), "Order buyAmount and sellAmount cannot be empty"

        url = self.base_url + "orders/"
        order["feeAmount"] = order["feeAmount"] or self.get_fee(order)
        signable_bytes = order.signable_bytes(domain=self.domain_separator)
        signable_hash = Web3.keccak(signable_bytes)
        message = encode_defunct(primitive=signable_hash)
        signed_message = Account.from_key(private_key).sign_message(message)

        data_json = {
            "sellToken": order["sellToken"].lower(),
            "buyToken": order["buyToken"].lower(),
            "sellAmount": str(order["sellAmount"]),
            "buyAmount": str(order["buyAmount"]),
            "validTo": order["validTo"],
            "appData": HexBytes(order["appData"]).hex()
            if isinstance(order["appData"], bytes)
            else order["appData"],
            "feeAmount": str(order["feeAmount"]),
            "kind": order["kind"],
            "partiallyFillable": order["partiallyFillable"],
            "signature": signed_message.signature.hex(),
            "signingScheme": "ethsign",
            "from": Account.from_key(private_key).address,
        }
        r = requests.post(url, json=data_json)
        if r.ok:
            return HexStr(r.json())
        else:
            return ErrorResponse(r.json())

    def get_trades(
        self, order_ui: Optional[HexStr] = None, owner: Optional[ChecksumAddress] = None
    ) -> List[TradeResponse]:
        assert bool(order_ui) ^ bool(
            owner
        ), "order_ui or owner must be provided, but not both"
        url = self.base_url + "trades/?"
        if order_ui:
            url += f"orderUid={order_ui}"
        elif owner:
            url += f"owner={owner}"

        response = requests.get(url)
        if response.ok:
            return cast(List[TradeResponse], response.json())
        else:
            return []

    def get_estimated_amount(
        self,
        base_token: ChecksumAddress,
        quote_token: ChecksumAddress,
        kind: OrderKind,
        amount: int,
    ) -> Union[AmountResponse, ErrorResponse]:
        """
        The estimated amount in quote token for either buying or selling amount of baseToken.
        """
        url = self.base_url + f"markets/{base_token}-{quote_token}/{kind.name}/{amount}"
        response = requests.get(url)
        if response.ok:
            return AmountResponse(response.json())
        else:
            return ErrorResponse(response.json())



import dataclasses
import math
from enum import Enum
from logging import getLogger
from typing import Callable, List, NamedTuple, Optional, Union

from eth_account import Account
from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract
from web3.exceptions import BadFunctionCallOutput
from web3.types import BlockIdentifier, Wei

from gnosis.eth.constants import GAS_CALL_DATA_BYTE, NULL_ADDRESS, SENTINEL_ADDRESS
from gnosis.eth.contracts import (
    get_compatibility_fallback_handler_V1_3_0_contract,
    get_delegate_constructor_proxy_contract,
    get_safe_contract,
    get_safe_V0_0_1_contract,
    get_safe_V1_0_0_contract,
    get_safe_V1_1_1_contract,
    get_safe_V1_3_0_contract,
)
from gnosis.eth.ethereum_client import EthereumClient, EthereumTxSent
from gnosis.eth.utils import get_eth_address_with_key
from gnosis.safe.proxy_factory import ProxyFactory

from ..eth.typing import EthereumData
from .exceptions import (
    CannotEstimateGas,
    CannotRetrieveSafeInfoException,
    InvalidPaymentToken,
)
from .safe_create2_tx import SafeCreate2Tx, SafeCreate2TxBuilder
from .safe_creation_tx import InvalidERC20Token, SafeCreationTx
from .safe_tx import SafeTx

try:
    from functools import cache
except ImportError:
    from functools import lru_cache

    cache = lru_cache(maxsize=None)


logger = getLogger(__name__)


class SafeCreationEstimate(NamedTuple):
    gas: int
    gas_price: int
    payment: int
    payment_token: Optional[str]


class SafeOperation(Enum):
    CALL = 0
    DELEGATE_CALL = 1
    CREATE = 2


@dataclasses.dataclass
class SafeInfo:
    address: ChecksumAddress
    fallback_handler: ChecksumAddress
    guard: ChecksumAddress
    master_copy: ChecksumAddress
    modules: List[ChecksumAddress]
    nonce: int
    owners: List[ChecksumAddress]
    threshold: int
    version: str


class Safe:
    """
    Class to manage a Gnosis Safe
    """

    FALLBACK_HANDLER_STORAGE_SLOT = (
        0x6C9A6C4A39284E37ED1CF53D337577D14212A4870FB976A4366C693B939918D5
    )
    GUARD_STORAGE_SLOT = (
        0x4A204F620C8C5CCDCA3FD54D003BADD85BA500436A431F0CBDA4F558C93C34C8
    )

    def __init__(self, address: ChecksumAddress, ethereum_client: EthereumClient):
        """
        :param address: Safe address
        :param ethereum_client: Initialized ethereum client
        """
        assert Web3.isChecksumAddress(address), "%s is not a valid address" % address

        self.ethereum_client = ethereum_client
        self.w3 = self.ethereum_client.w3
        self.address = address

    def __str__(self):
        return f"Safe={self.address}"

    @staticmethod
    def create(
        ethereum_client: EthereumClient,
        deployer_account: LocalAccount,
        master_copy_address: str,
        owners: List[str],
        threshold: int,
        fallback_handler: str = NULL_ADDRESS,
        proxy_factory_address: Optional[str] = None,
        payment_token: str = NULL_ADDRESS,
        payment: int = 0,
        payment_receiver: str = NULL_ADDRESS,
    ) -> EthereumTxSent:

        """
        Deploy new Safe proxy pointing to the specified `master_copy` address and configured
        with the provided `owners` and `threshold`. By default, payment for the deployer of the tx will be `0`.
        If `proxy_factory_address` is set deployment will be done using the proxy factory instead of calling
        the `constructor` of a new `DelegatedProxy`
        Using `proxy_factory_address` is recommended, as it takes less gas.
        (Testing with `Ganache` and 1 owner 261534 without proxy vs 229022 with Proxy)
        """

        assert owners, "At least one owner must be set"
        assert threshold >= len(owners), "Threshold=%d must be >= %d" % (
            threshold,
            len(owners),
        )

        initializer = (
            get_safe_contract(ethereum_client.w3, NULL_ADDRESS)
            .functions.setup(
                owners,
                threshold,
                NULL_ADDRESS,  # Contract address for optional delegate call
                b"",  # Data payload for optional delegate call
                fallback_handler,  # Handler for fallback calls to this contract,
                payment_token,
                payment,
                payment_receiver,
            )
            .buildTransaction({"gas": Wei(1), "gasPrice": Wei(1)})["data"]
        )

        if proxy_factory_address:
            proxy_factory = ProxyFactory(proxy_factory_address, ethereum_client)
            return proxy_factory.deploy_proxy_contract(
                deployer_account, master_copy_address, initializer=HexBytes(initializer)
            )

        proxy_contract = get_delegate_constructor_proxy_contract(ethereum_client.w3)
        tx = proxy_contract.constructor(
            master_copy_address, initializer
        ).buildTransaction({"from": deployer_account.address})
        tx["gas"] = tx["gas"] * 100000
        tx_hash = ethereum_client.send_unsigned_transaction(
            tx, private_key=deployer_account.key
        )
        tx_receipt = ethereum_client.get_transaction_receipt(tx_hash, timeout=60)
        assert tx_receipt
        assert tx_receipt["status"]

        contract_address = tx_receipt["contractAddress"]
        return EthereumTxSent(tx_hash, tx, contract_address)

    @staticmethod
    def _deploy_master_contract(
        ethereum_client: EthereumClient,
        deployer_account: LocalAccount,
        contract_fn: Callable[[Web3, Optional[str]], Contract],
    ) -> EthereumTxSent:
        """
        Deploy master contract. Takes deployer_account (if unlocked in the node) or the deployer private key
        Safe with version > v1.1.1 doesn't need to be initialized as it already has a constructor

        :param ethereum_client:
        :param deployer_account: Ethereum account
        :param contract_fn: get contract function
        :return: deployed contract address
        """
        safe_contract = contract_fn(ethereum_client.w3)
        constructor_tx = safe_contract.constructor().buildTransaction()
        tx_hash = ethereum_client.send_unsigned_transaction(
            constructor_tx, private_key=deployer_account.key
        )
        tx_receipt = ethereum_client.get_transaction_receipt(tx_hash, timeout=60)
        assert tx_receipt
        assert tx_receipt["status"]

        ethereum_tx_sent = EthereumTxSent(
            tx_hash, constructor_tx, tx_receipt["contractAddress"]
        )
        logger.info(
            "Deployed and initialized Safe Master Contract version=%s on address %s by %s",
            contract_fn(ethereum_client.w3, ethereum_tx_sent.contract_address)
            .functions.VERSION()
            .call(),
            ethereum_tx_sent.contract_address,
            deployer_account.address,
        )
        return ethereum_tx_sent

    @classmethod
    def deploy_compatibility_fallback_handler(
        cls, ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy Compatibility Fallback handler v1.3.0

        :param ethereum_client:
        :param deployer_account: Ethereum account
        :return: deployed contract address
        """

        contract = get_compatibility_fallback_handler_V1_3_0_contract(
            ethereum_client.w3
        )
        constructor_tx = contract.constructor().buildTransaction()
        tx_hash = ethereum_client.send_unsigned_transaction(
            constructor_tx, private_key=deployer_account.key
        )
        tx_receipt = ethereum_client.get_transaction_receipt(tx_hash, timeout=60)
        assert tx_receipt
        assert tx_receipt["status"]

        ethereum_tx_sent = EthereumTxSent(
            tx_hash, constructor_tx, tx_receipt["contractAddress"]
        )
        logger.info(
            "Deployed and initialized Compatibility Fallback Handler version=%s on address %s by %s",
            "1.3.0",
            ethereum_tx_sent.contract_address,
            deployer_account.address,
        )
        return ethereum_tx_sent

    @classmethod
    def deploy_master_contract_v1_3_0(
        cls, ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy master contract v1.3.0. Takes deployer_account (if unlocked in the node) or the deployer private key
        Safe with version > v1.1.1 doesn't need to be initialized as it already has a constructor

        :param ethereum_client:
        :param deployer_account: Ethereum account
        :return: deployed contract address
        """

        return cls._deploy_master_contract(
            ethereum_client, deployer_account, get_safe_V1_3_0_contract
        )

    @classmethod
    def deploy_master_contract_v1_1_1(
        cls, ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy master contract v1.1.1. Takes deployer_account (if unlocked in the node) or the deployer private key
        Safe with version > v1.1.1 doesn't need to be initialized as it already has a constructor

        :param ethereum_client:
        :param deployer_account: Ethereum account
        :return: deployed contract address
        """

        return cls._deploy_master_contract(
            ethereum_client, deployer_account, get_safe_V1_1_1_contract
        )

    @staticmethod
    def deploy_master_contract_v1_0_0(
        ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy master contract. Takes deployer_account (if unlocked in the node) or the deployer private key

        :param ethereum_client:
        :param deployer_account: Ethereum account
        :return: deployed contract address
        """

        safe_contract = get_safe_V1_0_0_contract(ethereum_client.w3)
        constructor_data = safe_contract.constructor().buildTransaction({"gas": 0})[
            "data"
        ]
        initializer_data = safe_contract.functions.setup(
            # We use 2 owners that nobody controls for the master copy
            [
                "0x0000000000000000000000000000000000000002",
                "0x0000000000000000000000000000000000000003",
            ],
            2,  # Threshold. Maximum security
            NULL_ADDRESS,  # Address for optional DELEGATE CALL
            b"",  # Data for optional DELEGATE CALL
            NULL_ADDRESS,  # Payment token
            0,  # Payment
            NULL_ADDRESS,  # Refund receiver
        ).buildTransaction({"to": NULL_ADDRESS})["data"]

        ethereum_tx_sent = ethereum_client.deploy_and_initialize_contract(
            deployer_account, constructor_data, HexBytes(initializer_data)
        )
        logger.info(
            "Deployed and initialized Safe Master Contract=%s by %s",
            ethereum_tx_sent.contract_address,
            deployer_account.address,
        )
        return ethereum_tx_sent

    @staticmethod
    def deploy_master_contract_v0_0_1(
        ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy master contract. Takes deployer_account (if unlocked in the node) or the deployer private key

        :param ethereum_client:
        :param deployer_account: Ethereum account
        :return: deployed contract address
        """

        safe_contract = get_safe_V0_0_1_contract(ethereum_client.w3)
        constructor_data = safe_contract.constructor().buildTransaction({"gas": 0})[
            "data"
        ]
        initializer_data = safe_contract.functions.setup(
            # We use 2 owners that nobody controls for the master copy
            [
                "0x0000000000000000000000000000000000000002",
                "0x0000000000000000000000000000000000000003",
            ],
            2,  # Threshold. Maximum security
            NULL_ADDRESS,  # Address for optional DELEGATE CALL
            b"",  # Data for optional DELEGATE CALL
        ).buildTransaction({"to": NULL_ADDRESS})["data"]

        ethereum_tx_sent = ethereum_client.deploy_and_initialize_contract(
            deployer_account, constructor_data, HexBytes(initializer_data)
        )
        logger.info(
            "Deployed and initialized Old Safe Master Contract=%s by %s",
            ethereum_tx_sent.contract_address,
            deployer_account.address,
        )
        return ethereum_tx_sent

    @staticmethod
    def estimate_safe_creation(
        ethereum_client: EthereumClient,
        old_master_copy_address: str,
        number_owners: int,
        gas_price: int,
        payment_token: Optional[str],
        payment_receiver: str = NULL_ADDRESS,
        payment_token_eth_value: float = 1.0,
        fixed_creation_cost: Optional[int] = None,
    ) -> SafeCreationEstimate:
        s = 15
        owners = [get_eth_address_with_key()[0] for _ in range(number_owners)]
        threshold = number_owners
        safe_creation_tx = SafeCreationTx(
            w3=ethereum_client.w3,
            owners=owners,
            threshold=threshold,
            signature_s=s,
            master_copy=old_master_copy_address,
            gas_price=gas_price,
            funder=payment_receiver,
            payment_token=payment_token,
            payment_token_eth_value=payment_token_eth_value,
            fixed_creation_cost=fixed_creation_cost,
        )
        return SafeCreationEstimate(
            safe_creation_tx.gas,
            safe_creation_tx.gas_price,
            safe_creation_tx.payment,
            safe_creation_tx.payment_token,
        )

    @staticmethod
    def estimate_safe_creation_2(
        ethereum_client: EthereumClient,
        master_copy_address: str,
        proxy_factory_address: str,
        number_owners: int,
        gas_price: int,
        payment_token: Optional[str],
        payment_receiver: str = NULL_ADDRESS,
        fallback_handler: Optional[str] = None,
        payment_token_eth_value: float = 1.0,
        fixed_creation_cost: Optional[int] = None,
    ) -> SafeCreationEstimate:
        salt_nonce = 15
        owners = [Account.create().address for _ in range(number_owners)]
        threshold = number_owners
        if not fallback_handler:
            fallback_handler = (
                Account.create().address
            )  # Better estimate it, it's required for new Safes
        safe_creation_tx = SafeCreate2TxBuilder(
            w3=ethereum_client.w3,
            master_copy_address=master_copy_address,
            proxy_factory_address=proxy_factory_address,
        ).build(
            owners=owners,
            threshold=threshold,
            fallback_handler=fallback_handler,
            salt_nonce=salt_nonce,
            gas_price=gas_price,
            payment_receiver=payment_receiver,
            payment_token=payment_token,
            payment_token_eth_value=payment_token_eth_value,
            fixed_creation_cost=fixed_creation_cost,
        )
        return SafeCreationEstimate(
            safe_creation_tx.gas,
            safe_creation_tx.gas_price,
            safe_creation_tx.payment,
            safe_creation_tx.payment_token,
        )

    @staticmethod
    def build_safe_creation_tx(
        ethereum_client: EthereumClient,
        master_copy_old_address: str,
        s: int,
        owners: List[str],
        threshold: int,
        gas_price: int,
        payment_token: Optional[str],
        payment_receiver: str,
        payment_token_eth_value: float = 1.0,
        fixed_creation_cost: Optional[int] = None,
    ) -> SafeCreationTx:
        try:
            safe_creation_tx = SafeCreationTx(
                w3=ethereum_client.w3,
                owners=owners,
                threshold=threshold,
                signature_s=s,
                master_copy=master_copy_old_address,
                gas_price=gas_price,
                funder=payment_receiver,
                payment_token=payment_token,
                payment_token_eth_value=payment_token_eth_value,
                fixed_creation_cost=fixed_creation_cost,
            )
        except InvalidERC20Token as exc:
            raise InvalidPaymentToken(
                "Invalid payment token %s" % payment_token
            ) from exc

        assert safe_creation_tx.tx_pyethereum.nonce == 0
        return safe_creation_tx

    @staticmethod
    def build_safe_create2_tx(
        ethereum_client: EthereumClient,
        master_copy_address: str,
        proxy_factory_address: str,
        salt_nonce: int,
        owners: List[str],
        threshold: int,
        gas_price: int,
        payment_token: Optional[str],
        payment_receiver: Optional[str] = None,  # If none, it will be `tx.origin`
        fallback_handler: Optional[str] = NULL_ADDRESS,
        payment_token_eth_value: float = 1.0,
        fixed_creation_cost: Optional[int] = None,
    ) -> SafeCreate2Tx:
        """
        Prepare safe proxy deployment for being relayed. It calculates and sets the costs of deployment to be returned
        to the sender of the tx. If you are an advanced user you may prefer to use `create` function
        """
        try:
            safe_creation_tx = SafeCreate2TxBuilder(
                w3=ethereum_client.w3,
                master_copy_address=master_copy_address,
                proxy_factory_address=proxy_factory_address,
            ).build(
                owners=owners,
                threshold=threshold,
                fallback_handler=fallback_handler,
                salt_nonce=salt_nonce,
                gas_price=gas_price,
                payment_receiver=payment_receiver,
                payment_token=payment_token,
                payment_token_eth_value=payment_token_eth_value,
                fixed_creation_cost=fixed_creation_cost,
            )
        except InvalidERC20Token as exc:
            raise InvalidPaymentToken(
                "Invalid payment token %s" % payment_token
            ) from exc

        return safe_creation_tx

    def check_funds_for_tx_gas(
        self, safe_tx_gas: int, base_gas: int, gas_price: int, gas_token: str
    ) -> bool:
        """
        Check safe has enough funds to pay for a tx

        :param safe_tx_gas: Safe tx gas
        :param base_gas: Data gas
        :param gas_price: Gas Price
        :param gas_token: Gas Token, to use token instead of ether for the gas
        :return: `True` if enough funds, `False` otherwise
        """
        if gas_token == NULL_ADDRESS:
            balance = self.ethereum_client.get_balance(self.address)
        else:
            balance = self.ethereum_client.erc20.get_balance(self.address, gas_token)
        return balance >= (safe_tx_gas + base_gas) * gas_price

    def estimate_tx_base_gas(
        self,
        to: str,
        value: int,
        data: bytes,
        operation: int,
        gas_token: str,
        estimated_tx_gas: int,
    ) -> int:
        """
        Calculate gas costs that are independent of the transaction execution(e.g. base transaction fee,
        signature check, payment of the refund...)

        :param to:
        :param value:
        :param data:
        :param operation:
        :param gas_token:
        :param estimated_tx_gas: gas calculated with `estimate_tx_gas`
        :return:
        """
        data = data or b""
        safe_contract = self.get_contract()
        threshold = self.retrieve_threshold()
        nonce = self.retrieve_nonce()

        # Every byte == 0 -> 4  Gas
        # Every byte != 0 -> 16 Gas (68 before Istanbul)
        # numbers < 256 (0x00(31*2)..ff) are 192 -> 31 * 4 + 1 * GAS_CALL_DATA_BYTE
        # numbers < 65535 (0x(30*2)..ffff) are 256 -> 30 * 4 + 2 * GAS_CALL_DATA_BYTE

        # Calculate gas for signatures
        # (array count (3 -> r, s, v) + ecrecover costs) * signature count
        # ecrecover for ecdsa ~= 4K gas, we use 6K
        ecrecover_gas = 6000
        signature_gas = threshold * (
            1 * GAS_CALL_DATA_BYTE + 2 * 32 * GAS_CALL_DATA_BYTE + ecrecover_gas
        )

        safe_tx_gas = estimated_tx_gas
        base_gas = 0
        gas_price = 1
        gas_token = gas_token or NULL_ADDRESS
        signatures = b""
        refund_receiver = NULL_ADDRESS
        data = HexBytes(
            safe_contract.functions.execTransaction(
                to,
                value,
                data,
                operation,
                safe_tx_gas,
                base_gas,
                gas_price,
                gas_token,
                refund_receiver,
                signatures,
            ).buildTransaction({"gas": 1, "gasPrice": 1})["data"]
        )

        # If nonce == 0, nonce storage has to be initialized
        if nonce == 0:
            nonce_gas = 20000
        else:
            nonce_gas = 5000

        # Keccak costs for the hash of the safe tx
        hash_generation_gas = 1500

        base_gas = (
            signature_gas
            + self.ethereum_client.estimate_data_gas(data)
            + nonce_gas
            + hash_generation_gas
        )

        # Add additional gas costs
        if base_gas > 65536:
            base_gas += 64
        else:
            base_gas += 128

        base_gas += 32000  # Base tx costs, transfer costs...
        return base_gas

    def estimate_tx_gas_with_safe(
        self,
        to: str,
        value: int,
        data: bytes,
        operation: int,
        gas_limit: Optional[int] = None,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> int:
        """
        Estimate tx gas using safe `requiredTxGas` method

        :return: int: Estimated gas
        :raises: CannotEstimateGas: If gas cannot be estimated
        :raises: ValueError: Cannot decode received data
        """

        safe_address = self.address
        data = data or b""

        def parse_revert_data(result: bytes) -> int:
            # 4 bytes - error method id
            # 32 bytes - position
            # 32 bytes - length
            # Last 32 bytes - value of revert (if everything went right)
            gas_estimation_offset = 4 + 32 + 32
            gas_estimation = result[gas_estimation_offset:]

            # Estimated gas must be 32 bytes
            if len(gas_estimation) != 32:
                gas_limit_text = (
                    f"with gas limit={gas_limit} "
                    if gas_limit is not None
                    else "without gas limit set "
                )
                logger.warning(
                    "Safe=%s Problem estimating gas, returned value %sis %s for tx=%s",
                    safe_address,
                    gas_limit_text,
                    result.hex(),
                    tx,
                )
                raise CannotEstimateGas("Received %s for tx=%s" % (result.hex(), tx))

            return int(gas_estimation.hex(), 16)

        tx = (
            self.get_contract()
            .functions.requiredTxGas(to, value, data, operation)
            .buildTransaction(
                {
                    "from": safe_address,
                    "gas": 0,  # Don't call estimate
                    "gasPrice": 0,  # Don't get gas price
                }
            )
        )

        tx_params = {
            "from": safe_address,
            "to": safe_address,
            "data": tx["data"],
        }

        if gas_limit:
            tx_params["gas"] = hex(gas_limit)

        query = {
            "jsonrpc": "2.0",
            "method": "eth_call",
            "params": [tx_params, block_identifier],
            "id": 1,
        }

        response = self.ethereum_client.http_session.post(
            self.ethereum_client.ethereum_node_url, json=query, timeout=30
        )
        if response.ok:
            response_data = response.json()
            error_data: Optional[str] = None
            if "error" in response_data and "data" in response_data["error"]:
                error_data = response_data["error"]["data"]
            elif "result" in response_data:  # Ganache-cli
                error_data = response_data["result"]

            if error_data:
                if "0x" in error_data:
                    return parse_revert_data(
                        HexBytes(error_data[error_data.find("0x") :])
                    )

        raise CannotEstimateGas(
            f"Received {response.status_code} - {response.content} from ethereum node"
        )

    def estimate_tx_gas_with_web3(self, to: str, value: int, data: EthereumData) -> int:
        """
        :param to:
        :param value:
        :param data:
        :return: Estimation using web3 `estimate_gas`
        """
        try:
            return self.ethereum_client.estimate_gas(
                to, from_=self.address, value=value, data=data
            )
        except ValueError as exc:
            raise CannotEstimateGas(
                f"Cannot estimate gas with `eth_estimateGas`: {exc}"
            ) from exc

    def estimate_tx_gas_by_trying(
        self, to: str, value: int, data: Union[bytes, str], operation: int
    ):
        """
        Try to get an estimation with Safe's `requiredTxGas`. If estimation if successful, try to set a gas limit and
        estimate again. If gas estimation is ok, same gas estimation should be returned, if it's less than required
        estimation will not be completed, so estimation was not accurate and gas limit needs to be increased.

        :param to:
        :param value:
        :param data:
        :param operation:
        :return: Estimated gas calling `requiredTxGas` setting a gas limit and checking if `eth_call` is successful
        :raises: CannotEstimateGas
        """
        if not data:
            data = b""
        elif isinstance(data, str):
            data = HexBytes(data)

        gas_estimated = self.estimate_tx_gas_with_safe(to, value, data, operation)
        block_gas_limit: Optional[int] = None
        base_gas: Optional[int] = self.ethereum_client.estimate_data_gas(data)

        for i in range(
            1, 30
        ):  # Make sure tx can be executed, fixing for example 63/64th problem
            try:
                self.estimate_tx_gas_with_safe(
                    to,
                    value,
                    data,
                    operation,
                    gas_limit=gas_estimated + base_gas + 32000,
                )
                return gas_estimated
            except CannotEstimateGas:
                logger.warning(
                    "Safe=%s - Found 63/64 problem gas-estimated=%d to=%s data=%s",
                    self.address,
                    gas_estimated,
                    to,
                    data.hex(),
                )
                block_gas_limit = (
                    block_gas_limit
                    or self.w3.eth.get_block("latest", full_transactions=False)[
                        "gasLimit"
                    ]
                )
                gas_estimated = math.floor((1 + i * 0.03) * gas_estimated)
                if gas_estimated >= block_gas_limit:
                    return block_gas_limit
        return gas_estimated

    def estimate_tx_gas(self, to: str, value: int, data: bytes, operation: int) -> int:
        """
        Estimate tx gas. Use `requiredTxGas` on the Safe contract and fallbacks to `eth_estimateGas` if that method
        fails. Note: `eth_estimateGas` cannot estimate delegate calls

        :param to:
        :param value:
        :param data:
        :param operation:
        :return: Estimated gas for Safe inner tx
        :raises: CannotEstimateGas
        """
        # Costs to route through the proxy and nested calls
        PROXY_GAS = 1000
        # https://github.com/ethereum/solidity/blob/dfe3193c7382c80f1814247a162663a97c3f5e67/libsolidity/codegen/ExpressionCompiler.cpp#L1764
        # This was `false` before solc 0.4.21 -> `m_context.evmVersion().canOverchargeGasForCall()`
        # So gas needed by caller will be around 35k
        OLD_CALL_GAS = 35000
        # Web3 `estimate_gas` estimates less gas
        WEB3_ESTIMATION_OFFSET = 20000
        ADDITIONAL_GAS = PROXY_GAS + OLD_CALL_GAS

        try:
            return (
                self.estimate_tx_gas_by_trying(to, value, data, operation)
                + ADDITIONAL_GAS
            )
        except CannotEstimateGas:
            return (
                self.estimate_tx_gas_with_web3(to, value, data)
                + ADDITIONAL_GAS
                + WEB3_ESTIMATION_OFFSET
            )

    def estimate_tx_operational_gas(self, data_bytes_length: int):
        """
        DEPRECATED. `estimate_tx_base_gas` already includes this.
        Estimates the gas for the verification of the signatures and other safe related tasks
        before and after executing a transaction.
        Calculation will be the sum of:

            - Base cost of 15000 gas
            - 100 of gas per word of `data_bytes`
            - Validate the signatures 5000 * threshold (ecrecover for ecdsa ~= 4K gas)

        :param data_bytes_length: Length of the data (in bytes, so `len(HexBytes('0x12'))` would be `1`
        :return: gas costs per signature * threshold of Safe
        """
        threshold = self.retrieve_threshold()
        return 15000 + data_bytes_length // 32 * 100 + 5000 * threshold

    @cache
    def get_contract(self) -> Contract:
        v_1_3_0_contract = get_safe_V1_3_0_contract(self.w3, address=self.address)
        version = v_1_3_0_contract.functions.VERSION().call()
        if version == "1.3.0":
            return v_1_3_0_contract
        else:
            return get_safe_V1_1_1_contract(self.w3, address=self.address)

    def retrieve_all_info(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> SafeInfo:
        """
        Get all Safe info in the same batch call.

        :param block_identifier:
        :return:
        :raises: CannotRetrieveSafeInfoException
        """
        try:
            contract = self.get_contract()
            master_copy = self.retrieve_master_copy_address()
            fallback_handler = self.retrieve_fallback_handler()
            guard = self.retrieve_guard()

            results = self.ethereum_client.batch_call(
                [
                    contract.functions.getModulesPaginated(
                        SENTINEL_ADDRESS, 20
                    ),  # Does not exist in version < 1.1.1
                    contract.functions.nonce(),
                    contract.functions.getOwners(),
                    contract.functions.getThreshold(),
                    contract.functions.VERSION(),
                ],
                from_address=self.address,
                block_identifier=block_identifier,
                raise_exception=False,
            )
            modules_response, nonce, owners, threshold, version = results
            if modules_response:
                modules, next_module = modules_response
            if (
                not modules_response or next_module != SENTINEL_ADDRESS
            ):  # < 1.1.1 or still more elements in the list
                modules = self.retrieve_modules()
            return SafeInfo(
                self.address,
                fallback_handler,
                guard,
                master_copy,
                modules,
                nonce,
                owners,
                threshold,
                version,
            )
        except (ValueError, BadFunctionCallOutput) as e:
            raise CannotRetrieveSafeInfoException(self.address) from e

    def retrieve_code(self) -> HexBytes:
        return self.w3.eth.get_code(self.address)

    def retrieve_fallback_handler(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> ChecksumAddress:
        address = self.ethereum_client.w3.eth.get_storage_at(
            self.address,
            self.FALLBACK_HANDLER_STORAGE_SLOT,
            block_identifier=block_identifier,
        )[-20:]
        if len(address) == 20:
            return Web3.toChecksumAddress(address)
        else:
            return NULL_ADDRESS

    def retrieve_guard(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> ChecksumAddress:
        address = self.ethereum_client.w3.eth.get_storage_at(
            self.address, self.GUARD_STORAGE_SLOT, block_identifier=block_identifier
        )[-20:]
        if len(address) == 20:
            return Web3.toChecksumAddress(address)
        else:
            return NULL_ADDRESS

    def retrieve_master_copy_address(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> ChecksumAddress:
        bytes_address = self.w3.eth.get_storage_at(
            self.address, "0x00", block_identifier=block_identifier
        )[-20:]
        int_address = int.from_bytes(bytes_address, byteorder="big")
        return Web3.toChecksumAddress("{:#042x}".format(int_address))

    def retrieve_modules(
        self,
        pagination: Optional[int] = 50,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> List[str]:
        """
        :param pagination: Number of modules to get per request
        :param block_identifier:
        :return: List of module addresses
        """
        try:
            # Contracts with Safe version < 1.1.0 were not paginated
            contract = get_safe_V1_0_0_contract(
                self.ethereum_client.w3, address=self.address
            )
            return contract.functions.getModules().call(
                block_identifier=block_identifier
            )
        except BadFunctionCallOutput:
            pass

        contract = self.get_contract()
        address = SENTINEL_ADDRESS
        all_modules: List[str] = []
        while True:
            (modules, address) = contract.functions.getModulesPaginated(
                address, pagination
            ).call(block_identifier=block_identifier)

            all_modules.extend(modules)
            if address == SENTINEL_ADDRESS:
                break
            else:
                all_modules.append(address)
        return all_modules

    def retrieve_is_hash_approved(
        self,
        owner: str,
        safe_hash: bytes,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> bool:
        return (
            self.get_contract()
            .functions.approvedHashes(owner, safe_hash)
            .call(block_identifier=block_identifier)
            == 1
        )

    def retrieve_is_message_signed(
        self,
        message_hash: bytes,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> bool:
        return (
            self.get_contract()
            .functions.signedMessages(message_hash)
            .call(block_identifier=block_identifier)
        )

    def retrieve_is_owner(
        self, owner: str, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> bool:
        return (
            self.get_contract()
            .functions.isOwner(owner)
            .call(block_identifier=block_identifier)
        )

    def retrieve_nonce(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> int:
        return (
            self.get_contract()
            .functions.nonce()
            .call(block_identifier=block_identifier)
        )

    def retrieve_owners(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> List[str]:
        return (
            self.get_contract()
            .functions.getOwners()
            .call(block_identifier=block_identifier)
        )

    def retrieve_threshold(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> int:
        return (
            self.get_contract()
            .functions.getThreshold()
            .call(block_identifier=block_identifier)
        )

    def retrieve_version(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> str:
        return (
            self.get_contract()
            .functions.VERSION()
            .call(block_identifier=block_identifier)
        )

    def build_multisig_tx(
        self,
        to: str,
        value: int,
        data: bytes,
        operation: int = SafeOperation.CALL.value,
        safe_tx_gas: int = 0,
        base_gas: int = 0,
        gas_price: int = 0,
        gas_token: str = NULL_ADDRESS,
        refund_receiver: str = NULL_ADDRESS,
        signatures: bytes = b"",
        safe_nonce: Optional[int] = None,
        safe_version: Optional[str] = None,
    ) -> SafeTx:
        """
        Allows to execute a Safe transaction confirmed by required number of owners and then pays the account
        that submitted the transaction. The fees are always transfered, even if the user transaction fails

        :param to: Destination address of Safe transaction
        :param value: Ether value of Safe transaction
        :param data: Data payload of Safe transaction
        :param operation: Operation type of Safe transaction
        :param safe_tx_gas: Gas that should be used for the Safe transaction
        :param base_gas: Gas costs for that are independent of the transaction execution
            (e.g. base transaction fee, signature check, payment of the refund)
        :param gas_price: Gas price that should be used for the payment calculation
        :param gas_token: Token address (or `0x000..000` if ETH) that is used for the payment
        :param refund_receiver: Address of receiver of gas payment (or `0x000..000`  if tx.origin).
        :param signatures: Packed signature data ({bytes32 r}{bytes32 s}{uint8 v})
        :param safe_nonce: Nonce of the safe (to calculate hash)
        :param safe_version: Safe version (to calculate hash)
        :return:
        """

        if safe_nonce is None:
            safe_nonce = self.retrieve_nonce()
        safe_version = safe_version or self.retrieve_version()
        return SafeTx(
            self.ethereum_client,
            self.address,
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            base_gas,
            gas_price,
            gas_token,
            refund_receiver,
            signatures=signatures,
            safe_nonce=safe_nonce,
            safe_version=safe_version,
        )

    def send_multisig_tx(
        self,
        to: str,
        value: int,
        data: bytes,
        operation: int,
        safe_tx_gas: int,
        base_gas: int,
        gas_price: int,
        gas_token: str,
        refund_receiver: str,
        signatures: bytes,
        tx_sender_private_key: str,
        tx_gas=None,
        tx_gas_price=None,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> EthereumTxSent:
        """
        Build and send Safe tx

        :param to:
        :param value:
        :param data:
        :param operation:
        :param safe_tx_gas:
        :param base_gas:
        :param gas_price:
        :param gas_token:
        :param refund_receiver:
        :param signatures:
        :param tx_sender_private_key:
        :param tx_gas: Gas for the external tx. If not, `(safe_tx_gas + data_gas) * 2` will be used
        :param tx_gas_price: Gas price of the external tx. If not, `gas_price` will be used
        :param block_identifier:
        :return: Tuple(tx_hash, tx)
        :raises: InvalidMultisigTx: If user tx cannot go through the Safe
        """

        safe_tx = self.build_multisig_tx(
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            base_gas,
            gas_price,
            gas_token,
            refund_receiver,
            signatures,
        )

        tx_sender_address = Account.from_key(tx_sender_private_key).address
        safe_tx.call(
            tx_sender_address=tx_sender_address, block_identifier=block_identifier
        )

        tx_hash, tx = safe_tx.execute(
            tx_sender_private_key=tx_sender_private_key,
            tx_gas=tx_gas,
            tx_gas_price=tx_gas_price,
            block_identifier=block_identifier,
        )

        return EthereumTxSent(tx_hash, tx, None)



from django.core.exceptions import ValidationError
from django.test import TestCase

from ...utils import get_eth_address_with_key
from ..validators import validate_checksumed_address


class TestValidators(TestCase):
    def test_checksum_address_validator(self):
        eth_address, eth_key = get_eth_address_with_key()

        self.assertIsNone(validate_checksumed_address(eth_address))

        self.assertRaises(
            ValidationError, validate_checksumed_address, eth_address.lower()
        )



# https://etherscan.io/address/0x03403154afc09Ce8e44C3B185C82C6aD5f86b9ab
YVAULT_ABI = [
    {
        "inputs": [
            {"internalType": "address", "name": "_token", "type": "address"},
            {"internalType": "address", "name": "_controller", "type": "address"},
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "spender",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Approval",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Transfer",
        "type": "event",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "address", "name": "owner", "type": "address"},
            {"internalType": "address", "name": "spender", "type": "address"},
        ],
        "name": "allowance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "approve",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "available",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "balance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "account", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "controller",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "decimals",
        "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "subtractedValue", "type": "uint256"},
        ],
        "name": "decreaseAllowance",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "uint256", "name": "_amount", "type": "uint256"}],
        "name": "deposit",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [],
        "name": "depositAll",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [],
        "name": "earn",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "getPricePerFullShare",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "governance",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "addedValue", "type": "uint256"},
        ],
        "name": "increaseAllowance",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "max",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "min",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "name",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "_controller", "type": "address"}
        ],
        "name": "setController",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "_governance", "type": "address"}
        ],
        "name": "setGovernance",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "uint256", "name": "_min", "type": "uint256"}],
        "name": "setMin",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "string", "name": "_name", "type": "string"}],
        "name": "setName",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "string", "name": "_symbol", "type": "string"}],
        "name": "setSymbol",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "symbol",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "token",
        "outputs": [{"internalType": "contract IERC20", "name": "", "type": "address"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "transfer",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "sender", "type": "address"},
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "uint256", "name": "_shares", "type": "uint256"}],
        "name": "withdraw",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [],
        "name": "withdrawAll",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
]

YTOKEN_ABI = [
    {
        "name": "Transfer",
        "inputs": [
            {"name": "sender", "type": "address", "indexed": True},
            {"name": "receiver", "type": "address", "indexed": True},
            {"name": "value", "type": "uint256", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "Approval",
        "inputs": [
            {"name": "owner", "type": "address", "indexed": True},
            {"name": "spender", "type": "address", "indexed": True},
            {"name": "value", "type": "uint256", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "StrategyAdded",
        "inputs": [
            {"name": "strategy", "type": "address", "indexed": True},
            {"name": "debtRatio", "type": "uint256", "indexed": False},
            {"name": "minDebtPerHarvest", "type": "uint256", "indexed": False},
            {"name": "maxDebtPerHarvest", "type": "uint256", "indexed": False},
            {"name": "performanceFee", "type": "uint256", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "StrategyReported",
        "inputs": [
            {"name": "strategy", "type": "address", "indexed": True},
            {"name": "gain", "type": "uint256", "indexed": False},
            {"name": "loss", "type": "uint256", "indexed": False},
            {"name": "debtPaid", "type": "uint256", "indexed": False},
            {"name": "totalGain", "type": "uint256", "indexed": False},
            {"name": "totalLoss", "type": "uint256", "indexed": False},
            {"name": "totalDebt", "type": "uint256", "indexed": False},
            {"name": "debtAdded", "type": "uint256", "indexed": False},
            {"name": "debtRatio", "type": "uint256", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "UpdateGovernance",
        "inputs": [{"name": "governance", "type": "address", "indexed": False}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "UpdateManagement",
        "inputs": [{"name": "management", "type": "address", "indexed": False}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "UpdateGuestList",
        "inputs": [{"name": "guestList", "type": "address", "indexed": False}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "UpdateRewards",
        "inputs": [{"name": "rewards", "type": "address", "indexed": False}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "UpdateDepositLimit",
        "inputs": [{"name": "depositLimit", "type": "uint256", "indexed": False}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "UpdatePerformanceFee",
        "inputs": [{"name": "performanceFee", "type": "uint256", "indexed": False}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "UpdateManagementFee",
        "inputs": [{"name": "managementFee", "type": "uint256", "indexed": False}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "UpdateGuardian",
        "inputs": [{"name": "guardian", "type": "address", "indexed": False}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "EmergencyShutdown",
        "inputs": [{"name": "active", "type": "bool", "indexed": False}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "UpdateWithdrawalQueue",
        "inputs": [{"name": "queue", "type": "address[20]", "indexed": False}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "StrategyUpdateDebtRatio",
        "inputs": [
            {"name": "strategy", "type": "address", "indexed": True},
            {"name": "debtRatio", "type": "uint256", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "StrategyUpdateMinDebtPerHarvest",
        "inputs": [
            {"name": "strategy", "type": "address", "indexed": True},
            {"name": "minDebtPerHarvest", "type": "uint256", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "StrategyUpdateMaxDebtPerHarvest",
        "inputs": [
            {"name": "strategy", "type": "address", "indexed": True},
            {"name": "maxDebtPerHarvest", "type": "uint256", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "StrategyUpdatePerformanceFee",
        "inputs": [
            {"name": "strategy", "type": "address", "indexed": True},
            {"name": "performanceFee", "type": "uint256", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "StrategyMigrated",
        "inputs": [
            {"name": "oldVersion", "type": "address", "indexed": True},
            {"name": "newVersion", "type": "address", "indexed": True},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "StrategyRevoked",
        "inputs": [{"name": "strategy", "type": "address", "indexed": True}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "StrategyRemovedFromQueue",
        "inputs": [{"name": "strategy", "type": "address", "indexed": True}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "StrategyAddedToQueue",
        "inputs": [{"name": "strategy", "type": "address", "indexed": True}],
        "anonymous": False,
        "type": "event",
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "initialize",
        "inputs": [
            {"name": "token", "type": "address"},
            {"name": "governance", "type": "address"},
            {"name": "rewards", "type": "address"},
            {"name": "nameOverride", "type": "string"},
            {"name": "symbolOverride", "type": "string"},
        ],
        "outputs": [],
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "initialize",
        "inputs": [
            {"name": "token", "type": "address"},
            {"name": "governance", "type": "address"},
            {"name": "rewards", "type": "address"},
            {"name": "nameOverride", "type": "string"},
            {"name": "symbolOverride", "type": "string"},
            {"name": "guardian", "type": "address"},
        ],
        "outputs": [],
    },
    {
        "stateMutability": "pure",
        "type": "function",
        "name": "apiVersion",
        "inputs": [],
        "outputs": [{"name": "", "type": "string"}],
        "gas": 4546,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setName",
        "inputs": [{"name": "name", "type": "string"}],
        "outputs": [],
        "gas": 107044,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setSymbol",
        "inputs": [{"name": "symbol", "type": "string"}],
        "outputs": [],
        "gas": 71894,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setGovernance",
        "inputs": [{"name": "governance", "type": "address"}],
        "outputs": [],
        "gas": 36365,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "acceptGovernance",
        "inputs": [],
        "outputs": [],
        "gas": 37637,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setManagement",
        "inputs": [{"name": "management", "type": "address"}],
        "outputs": [],
        "gas": 37775,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setGuestList",
        "inputs": [{"name": "guestList", "type": "address"}],
        "outputs": [],
        "gas": 37805,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setRewards",
        "inputs": [{"name": "rewards", "type": "address"}],
        "outputs": [],
        "gas": 37835,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setLockedProfitDegration",
        "inputs": [{"name": "degration", "type": "uint256"}],
        "outputs": [],
        "gas": 36519,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setDepositLimit",
        "inputs": [{"name": "limit", "type": "uint256"}],
        "outputs": [],
        "gas": 37795,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setPerformanceFee",
        "inputs": [{"name": "fee", "type": "uint256"}],
        "outputs": [],
        "gas": 37929,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setManagementFee",
        "inputs": [{"name": "fee", "type": "uint256"}],
        "outputs": [],
        "gas": 37959,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setGuardian",
        "inputs": [{"name": "guardian", "type": "address"}],
        "outputs": [],
        "gas": 39203,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setEmergencyShutdown",
        "inputs": [{"name": "active", "type": "bool"}],
        "outputs": [],
        "gas": 39274,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "setWithdrawalQueue",
        "inputs": [{"name": "queue", "type": "address[20]"}],
        "outputs": [],
        "gas": 763950,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "transfer",
        "inputs": [
            {"name": "receiver", "type": "address"},
            {"name": "amount", "type": "uint256"},
        ],
        "outputs": [{"name": "", "type": "bool"}],
        "gas": 76768,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "transferFrom",
        "inputs": [
            {"name": "sender", "type": "address"},
            {"name": "receiver", "type": "address"},
            {"name": "amount", "type": "uint256"},
        ],
        "outputs": [{"name": "", "type": "bool"}],
        "gas": 116531,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "approve",
        "inputs": [
            {"name": "spender", "type": "address"},
            {"name": "amount", "type": "uint256"},
        ],
        "outputs": [{"name": "", "type": "bool"}],
        "gas": 38271,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "increaseAllowance",
        "inputs": [
            {"name": "spender", "type": "address"},
            {"name": "amount", "type": "uint256"},
        ],
        "outputs": [{"name": "", "type": "bool"}],
        "gas": 40312,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "decreaseAllowance",
        "inputs": [
            {"name": "spender", "type": "address"},
            {"name": "amount", "type": "uint256"},
        ],
        "outputs": [{"name": "", "type": "bool"}],
        "gas": 40336,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "permit",
        "inputs": [
            {"name": "owner", "type": "address"},
            {"name": "spender", "type": "address"},
            {"name": "amount", "type": "uint256"},
            {"name": "expiry", "type": "uint256"},
            {"name": "signature", "type": "bytes"},
        ],
        "outputs": [{"name": "", "type": "bool"}],
        "gas": 81264,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "totalAssets",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 4098,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "deposit",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "deposit",
        "inputs": [{"name": "_amount", "type": "uint256"}],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "deposit",
        "inputs": [
            {"name": "_amount", "type": "uint256"},
            {"name": "recipient", "type": "address"},
        ],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "maxAvailableShares",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 383839,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "withdraw",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "withdraw",
        "inputs": [{"name": "maxShares", "type": "uint256"}],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "withdraw",
        "inputs": [
            {"name": "maxShares", "type": "uint256"},
            {"name": "recipient", "type": "address"},
        ],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "withdraw",
        "inputs": [
            {"name": "maxShares", "type": "uint256"},
            {"name": "recipient", "type": "address"},
            {"name": "maxLoss", "type": "uint256"},
        ],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "pricePerShare",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 18195,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "addStrategy",
        "inputs": [
            {"name": "strategy", "type": "address"},
            {"name": "debtRatio", "type": "uint256"},
            {"name": "minDebtPerHarvest", "type": "uint256"},
            {"name": "maxDebtPerHarvest", "type": "uint256"},
            {"name": "performanceFee", "type": "uint256"},
        ],
        "outputs": [],
        "gas": 1485796,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "updateStrategyDebtRatio",
        "inputs": [
            {"name": "strategy", "type": "address"},
            {"name": "debtRatio", "type": "uint256"},
        ],
        "outputs": [],
        "gas": 115193,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "updateStrategyMinDebtPerHarvest",
        "inputs": [
            {"name": "strategy", "type": "address"},
            {"name": "minDebtPerHarvest", "type": "uint256"},
        ],
        "outputs": [],
        "gas": 42441,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "updateStrategyMaxDebtPerHarvest",
        "inputs": [
            {"name": "strategy", "type": "address"},
            {"name": "maxDebtPerHarvest", "type": "uint256"},
        ],
        "outputs": [],
        "gas": 42471,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "updateStrategyPerformanceFee",
        "inputs": [
            {"name": "strategy", "type": "address"},
            {"name": "performanceFee", "type": "uint256"},
        ],
        "outputs": [],
        "gas": 41251,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "migrateStrategy",
        "inputs": [
            {"name": "oldVersion", "type": "address"},
            {"name": "newVersion", "type": "address"},
        ],
        "outputs": [],
        "gas": 1141468,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "revokeStrategy",
        "inputs": [],
        "outputs": [],
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "revokeStrategy",
        "inputs": [{"name": "strategy", "type": "address"}],
        "outputs": [],
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "addStrategyToQueue",
        "inputs": [{"name": "strategy", "type": "address"}],
        "outputs": [],
        "gas": 1199804,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "removeStrategyFromQueue",
        "inputs": [{"name": "strategy", "type": "address"}],
        "outputs": [],
        "gas": 23088703,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "debtOutstanding",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "debtOutstanding",
        "inputs": [{"name": "strategy", "type": "address"}],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "creditAvailable",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "creditAvailable",
        "inputs": [{"name": "strategy", "type": "address"}],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "availableDepositLimit",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 9551,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "expectedReturn",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "expectedReturn",
        "inputs": [{"name": "strategy", "type": "address"}],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "report",
        "inputs": [
            {"name": "gain", "type": "uint256"},
            {"name": "loss", "type": "uint256"},
            {"name": "_debtPayment", "type": "uint256"},
        ],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 1015170,
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "sweep",
        "inputs": [{"name": "token", "type": "address"}],
        "outputs": [],
    },
    {
        "stateMutability": "nonpayable",
        "type": "function",
        "name": "sweep",
        "inputs": [
            {"name": "token", "type": "address"},
            {"name": "amount", "type": "uint256"},
        ],
        "outputs": [],
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "name",
        "inputs": [],
        "outputs": [{"name": "", "type": "string"}],
        "gas": 8750,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "symbol",
        "inputs": [],
        "outputs": [{"name": "", "type": "string"}],
        "gas": 7803,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "decimals",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2408,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "precisionFactor",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2438,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "balanceOf",
        "inputs": [{"name": "arg0", "type": "address"}],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2683,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "allowance",
        "inputs": [
            {"name": "arg0", "type": "address"},
            {"name": "arg1", "type": "address"},
        ],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2928,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "totalSupply",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2528,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "token",
        "inputs": [],
        "outputs": [{"name": "", "type": "address"}],
        "gas": 2558,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "governance",
        "inputs": [],
        "outputs": [{"name": "", "type": "address"}],
        "gas": 2588,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "management",
        "inputs": [],
        "outputs": [{"name": "", "type": "address"}],
        "gas": 2618,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "guardian",
        "inputs": [],
        "outputs": [{"name": "", "type": "address"}],
        "gas": 2648,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "guestList",
        "inputs": [],
        "outputs": [{"name": "", "type": "address"}],
        "gas": 2678,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "strategies",
        "inputs": [{"name": "arg0", "type": "address"}],
        "outputs": [
            {"name": "performanceFee", "type": "uint256"},
            {"name": "activation", "type": "uint256"},
            {"name": "debtRatio", "type": "uint256"},
            {"name": "minDebtPerHarvest", "type": "uint256"},
            {"name": "maxDebtPerHarvest", "type": "uint256"},
            {"name": "lastReport", "type": "uint256"},
            {"name": "totalDebt", "type": "uint256"},
            {"name": "totalGain", "type": "uint256"},
            {"name": "totalLoss", "type": "uint256"},
        ],
        "gas": 11031,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "withdrawalQueue",
        "inputs": [{"name": "arg0", "type": "uint256"}],
        "outputs": [{"name": "", "type": "address"}],
        "gas": 2847,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "emergencyShutdown",
        "inputs": [],
        "outputs": [{"name": "", "type": "bool"}],
        "gas": 2768,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "depositLimit",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2798,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "debtRatio",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2828,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "totalDebt",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2858,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "lastReport",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2888,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "activation",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2918,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "lockedProfit",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2948,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "lockedProfitDegration",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 2978,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "rewards",
        "inputs": [],
        "outputs": [{"name": "", "type": "address"}],
        "gas": 3008,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "managementFee",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 3038,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "performanceFee",
        "inputs": [],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 3068,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "nonces",
        "inputs": [{"name": "arg0", "type": "address"}],
        "outputs": [{"name": "", "type": "uint256"}],
        "gas": 3313,
    },
    {
        "stateMutability": "view",
        "type": "function",
        "name": "DOMAIN_SEPARATOR",
        "inputs": [],
        "outputs": [{"name": "", "type": "bytes32"}],
        "gas": 3128,
    },
]



import os
import random
from logging import getLogger

from eth.constants import SECPK1_N
from web3 import Web3

from gnosis.eth.tests.utils import send_tx

from ..safe_creation_tx import SafeCreationTx

logger = getLogger(__name__)


def generate_salt_nonce() -> int:
    return random.getrandbits(256) - 1


def generate_valid_s() -> int:
    while True:
        s = int(os.urandom(30).hex(), 16)
        if s <= (SECPK1_N // 2):
            return s


def deploy_safe(
    w3: Web3,
    safe_creation_tx: SafeCreationTx,
    funder: str,
    initial_funding_wei: int = 0,
    funder_account=None,
) -> str:
    if funder_account:
        send_tx(
            w3,
            {
                "to": safe_creation_tx.deployer_address,
                "value": safe_creation_tx.payment,
            },
            funder_account,
        )

        send_tx(
            w3,
            {
                "to": safe_creation_tx.safe_address,
                "value": safe_creation_tx.payment + initial_funding_wei,
            },
            funder_account,
        )
    else:
        w3.eth.wait_for_transaction_receipt(
            w3.eth.send_transaction(
                {
                    "from": funder,
                    "to": safe_creation_tx.deployer_address,
                    "value": safe_creation_tx.payment,
                }
            )
        )

        w3.eth.wait_for_transaction_receipt(
            w3.eth.send_transaction(
                {
                    "from": funder,
                    "to": safe_creation_tx.safe_address,
                    "value": safe_creation_tx.payment + initial_funding_wei,
                }
            )
        )

    tx_hash = w3.eth.send_raw_transaction(bytes(safe_creation_tx.tx_raw))
    tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    assert tx_receipt.contractAddress == safe_creation_tx.safe_address
    assert tx_receipt.status

    return safe_creation_tx.safe_address



import binascii
from typing import Optional, Union

from django.core import exceptions
from django.db import models
from django.utils.translation import gettext_lazy as _

from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address
from hexbytes import HexBytes

from .filters import EthereumAddressFieldForm, Keccak256FieldForm
from .validators import validate_checksumed_address

try:
    from django.db import DefaultConnectionProxy

    connection = DefaultConnectionProxy()
except ImportError:
    from django.db import connections

    connection = connections["default"]


class EthereumAddressField(models.CharField):
    default_validators = [validate_checksumed_address]
    description = "DEPRECATED. Use `EthereumAddressV2Field`. Ethereum address (EIP55)"
    default_error_messages = {
        "invalid": _('"%(value)s" value must be an EIP55 checksummed address.'),
    }

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = 42
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["max_length"]
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection):
        return self.to_python(value)

    def to_python(self, value):
        value = super().to_python(value)
        if value:
            try:
                return to_checksum_address(value)
            except ValueError:
                raise exceptions.ValidationError(
                    self.error_messages["invalid"],
                    code="invalid",
                    params={"value": value},
                )
        else:
            return value

    def get_prep_value(self, value):
        value = super().get_prep_value(value)
        return self.to_python(value)


class EthereumAddressV2Field(models.Field):
    default_validators = [validate_checksumed_address]
    description = "Ethereum address (EIP55)"
    default_error_messages = {
        "invalid": _('"%(value)s" value must be an EIP55 checksummed address.'),
    }

    def get_internal_type(self):
        return "BinaryField"

    def from_db_value(
        self, value: memoryview, expression, connection
    ) -> Optional[ChecksumAddress]:
        if value:
            return to_checksum_address(value.hex())

    def get_prep_value(self, value: ChecksumAddress) -> Optional[bytes]:
        if value:
            return HexBytes(self.to_python(value))

    def to_python(self, value) -> Optional[ChecksumAddress]:
        if value is not None:
            try:
                return to_checksum_address(value)
            except ValueError:
                raise exceptions.ValidationError(
                    self.error_messages["invalid"],
                    code="invalid",
                    params={"value": value},
                )

    def formfield(self, **kwargs):
        defaults = {
            "form_class": EthereumAddressFieldForm,
            "max_length": 2 + 40,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)


class Uint256Field(models.DecimalField):
    description = _("Ethereum uint256 number")
    """
    Field to store ethereum uint256 values. Uses Decimal db type without decimals to store
    in the database, but retrieve as `int` instead of `Decimal` (https://docs.python.org/3/library/decimal.html)
    """

    def __init__(self, *args, **kwargs):
        kwargs["max_digits"] = 79  # 2 ** 256 is 78 digits
        kwargs["decimal_places"] = 0
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["max_digits"]
        del kwargs["decimal_places"]
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return int(value)


class HexField(models.CharField):
    """
    Field to store hex values (without 0x). Returns hex with 0x prefix.

    On Database side a CharField is used.
    """

    description = "Stores a hex value into an CharField"

    def from_db_value(self, value, expression, connection):
        return self.to_python(value)

    def to_python(self, value):
        return value if value is None else HexBytes(value).hex()

    def get_prep_value(self, value):
        if value is None:
            return value
        elif isinstance(value, HexBytes):
            return value.hex()[
                2:
            ]  # HexBytes.hex() retrieves hexadecimal with '0x', remove it
        elif isinstance(value, bytes):
            return value.hex()  # bytes.hex() retrieves hexadecimal without '0x'
        else:  # str
            return HexBytes(value).hex()[2:]

    def formfield(self, **kwargs):
        # We need max_lenght + 2 on forms because of `0x`
        defaults = {"max_length": self.max_length + 2}
        # TODO: Handle multiple backends with different feature flags.
        if self.null and not connection.features.interprets_empty_strings_as_nulls:
            defaults["empty_value"] = None
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def clean(self, value, model_instance):
        value = self.to_python(value)
        self.validate(value, model_instance)
        # Validation didn't work because of `0x`
        self.run_validators(value[2:])
        return value


class Sha3HashField(HexField):
    description = "DEPRECATED. Use `Keccak256Field`"

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = 64
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["max_length"]
        return name, path, args, kwargs


class Keccak256Field(models.BinaryField):
    description = "Keccak256 hash stored as binary"
    default_error_messages = {
        "invalid": _('"%(value)s" hash must be a 32 bytes hexadecimal.'),
        "length": _('"%(value)s" hash must have exactly 32 bytes.'),
    }

    def _to_bytes(self, value) -> Optional[bytes]:
        if value is None:
            return
        else:
            try:
                result = HexBytes(value)
                if len(result) != 32:
                    raise exceptions.ValidationError(
                        self.error_messages["length"],
                        code="length",
                        params={"value": value},
                    )
                return result
            except (ValueError, binascii.Error):
                raise exceptions.ValidationError(
                    self.error_messages["invalid"],
                    code="invalid",
                    params={"value": value},
                )

    def from_db_value(
        self, value: memoryview, expression, connection
    ) -> Optional[bytes]:
        if value:
            return HexBytes(value.tobytes()).hex()

    def get_prep_value(self, value: Union[bytes, str]) -> Optional[bytes]:
        if value:
            return self._to_bytes(value)

    def to_python(self, value) -> Optional[str]:
        if value is not None:
            try:
                return self._to_bytes(value)
            except (ValueError, binascii.Error):
                raise exceptions.ValidationError(
                    self.error_messages["invalid"],
                    code="invalid",
                    params={"value": value},
                )

    def formfield(self, **kwargs):
        defaults = {
            "form_class": Keccak256FieldForm,
            "max_length": 2 + 64,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)



from enum import Enum

from eip712_structs import Address, Boolean, Bytes, EIP712Struct, String, Uint


class Order(EIP712Struct):
    sellToken = Address()
    buyToken = Address()
    receiver = Address()
    sellAmount = Uint(256)
    buyAmount = Uint(256)
    validTo = Uint(32)
    appData = Bytes(32)
    feeAmount = Uint(256)
    kind = String()  # `sell` or `buy`
    partiallyFillable = Boolean()
    sellTokenBalance = String()  # `erc20`, `external` or `internal`
    buyTokenBalance = String()  # `erc20` or `internal`


class OrderKind(Enum):
    BUY = 0
    SELL = 1



# flake8: noqa F401
from .blockscout_client import (
    BlockscoutClient,
    BlockscoutClientException,
    BlockScoutConfigurationProblem,
)
from .contract_metadata import ContractMetadata
from .etherscan_client import (
    EtherscanClient,
    EtherscanClientConfigurationProblem,
    EtherscanClientException,
    EtherscanRateLimitError,
)
from .sourcify import Sourcify



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "README.rst")) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

requirements = [
    "cached-property>=1.5; python_version < '3.8'",
    "eip712_structs",
    "packaging",
    "py-evm>=0.5.0a3",
    "typing-extensions>=3.10; python_version < '3.8'",
    "requests>=2",
    "web3>=5.23.0",
]

extras_require = {"django": ["django>=2", "django-filter>=2", "djangorestframework>=2"]}

setup(
    name="pm-dev-gnosis-py",
    version="1.0.7",
    packages=find_packages(),
    package_data={"gnosis": ["py.typed"]},
    install_requires=requirements,
    include_package_data=True,
    extras_require=extras_require,
    python_requires=">=3.7",
    license="MIT License",
    description="PMGnosisLibraries",
    long_description=README,
    url="https://github.com/gnosis/gnosis-py",
    author="Uxío",
    author_email="uxio@gnosis.pm",
    keywords=["ethereum", "web3", "django", "rest", "gnosis"],
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



import os

import pytest
import requests
from eth_account.signers.local import LocalAccount
from web3 import Web3
from web3.contract import Contract

from ..contracts import get_example_erc20_contract


def just_test_if_mainnet_node() -> str:
    mainnet_node_url = os.environ.get("ETHEREUM_MAINNET_NODE")
    if hasattr(just_test_if_mainnet_node, "checked"):  # Just check node first time
        return mainnet_node_url

    if not mainnet_node_url:
        pytest.skip(
            "Mainnet node not defined, cannot test oracles", allow_module_level=True
        )
    else:
        try:
            if not requests.post(
                mainnet_node_url,
                timeout=5,
                json={
                    "jsonrpc": "2.0",
                    "method": "eth_blockNumber",
                    "params": [],
                    "id": 1,
                },
            ).ok:
                pytest.skip("Cannot connect to mainnet node", allow_module_level=True)
        except IOError:
            pytest.skip(
                "Problem connecting to the mainnet node", allow_module_level=True
            )
    just_test_if_mainnet_node.checked = True
    return mainnet_node_url


def send_tx(w3: Web3, tx, account: LocalAccount) -> bytes:
    tx["from"] = account.address
    if "nonce" not in tx:
        tx["nonce"] = w3.eth.get_transaction_count(
            account.address, block_identifier="pending"
        )

    if "gasPrice" not in tx:
        tx["gasPrice"] = w3.eth.gas_price

    if "gas" not in tx:
        tx["gas"] = w3.eth.estimate_gas(tx)
    else:
        tx["gas"] *= 2

    signed_tx = account.sign_transaction(tx)
    tx_hash = w3.eth.send_raw_transaction(bytes(signed_tx.rawTransaction))
    tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    assert tx_receipt.status == 1, "Error with tx %s - %s" % (tx_hash.hex(), tx)
    return tx_hash


def deploy_example_erc20(
    w3: Web3,
    amount: int,
    owner: str,
    deployer: str = None,
    account: LocalAccount = None,
) -> Contract:
    return deploy_erc20(
        w3, "Uxio", "UXI", owner, amount, deployer=deployer, account=account
    )


def deploy_erc20(
    w3: Web3,
    name: str,
    symbol: str,
    owner: str,
    amount: int,
    decimals: int = 18,
    deployer: str = None,
    account: LocalAccount = None,
) -> Contract:
    if account:
        erc20_contract = get_example_erc20_contract(w3)
        tx = erc20_contract.constructor(
            name, symbol, decimals, owner, amount
        ).buildTransaction(
            {
                "nonce": w3.eth.get_transaction_count(
                    account.address, block_identifier="pending"
                )
            }
        )
        signed_tx = account.sign_transaction(tx)
        tx_hash = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    else:
        deployer = deployer or w3.eth.accounts[0]
        erc20_contract = get_example_erc20_contract(w3)
        tx_hash = erc20_contract.constructor(
            name, symbol, decimals, owner, amount
        ).transact({"from": deployer})

    tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    erc20_address = tx_receipt.contractAddress
    deployed_erc20 = get_example_erc20_contract(w3, erc20_address)
    assert deployed_erc20.functions.balanceOf(owner).call() == amount
    return deployed_erc20



ZERION_TOKEN_ADAPTER_ABI = [
    {
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "getComponents",
        "outputs": [
            {
                "components": [
                    {"internalType": "address", "name": "token", "type": "address"},
                    {"internalType": "string", "name": "tokenType", "type": "string"},
                    {"internalType": "uint256", "name": "rate", "type": "uint256"},
                ],
                "internalType": "struct Component[]",
                "name": "",
                "type": "tuple[]",
            }
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "getMetadata",
        "outputs": [
            {
                "components": [
                    {"internalType": "address", "name": "token", "type": "address"},
                    {"internalType": "string", "name": "name", "type": "string"},
                    {"internalType": "string", "name": "symbol", "type": "string"},
                    {"internalType": "uint8", "name": "decimals", "type": "uint8"},
                ],
                "internalType": "struct TokenMetadata",
                "name": "",
                "type": "tuple",
            }
        ],
        "stateMutability": "view",
        "type": "function",
    },
]
